 (**********************************************)
 (*                                            *)
 (*   Dynamic List v1.33 for Delphi 3/4/5      *)
 (*                                            *)
 (*   Copiright 1999 by George Barbakadze      *)
 (*   All rights reserved                      *)
 (*                                            *)
 (*   This component is freeware               *)
 (**********************************************)

{             layout of characters in memory


                /\                                     /\
 |--------------|-------------TSymbolGroup-------------|--------------|
 |  |-----------|-------------|           |------------|-----------|  |
 |  | PrevGroup: PSymbolGroup |           |     PrevChar: char     |  |
 |  |-------------------------|           |------------------------|  |
 |                                                                    |
 |  |-------------------array[char] of TSymbol---------------------|  |
 |  |---------TSymbol---------|    |------|   |------|      |------|  |
 |  |            #0           |    |  #1  |   |  #2  |      | #255 |  |
 |  | Active: boolean         |    |      |   |      |      |      |  |
 |  | Structure: integer      |    |      |   |      | ...  |      |  |
 |  | Additional: pointer     |<-| |      |   |      |      |      |  |
 |  | ----------------------- |  | |      |   |      |      |      |  |
 |  | NextGroup: PSymbolGroup |  | |      |   |      |      |      |  |
 |  |-----|-------------------|  | |------|   |------|      |------|  |
 |--------|----------------------|------------------------------------|
          |     /\               |
          |     |                ----------------------|
         \/     |                                      |
 |--------------|-------------TSymbolGroup-------------|--------------|
 |  |-----------|-------------|           |------------|-----------|  |
 |  | PrevGroup: PSymbolGroup |           |     PrevChar: char     |  |
 |  |-------------------------|           |------------------------|  |
 |                                                                    |
 |  |-------------------array[char] of TSymbol---------------------|  |
 |  |---------TSymbol---------|    |------|   |------|      |------|  |
 |  |            #0           |    |  #1  |   |  #2  |      | #255 |  |
 |  | Active: boolean         |    |      |   |      |      |      |  |
 |  | Structure: integer      |    |      |   |      | ...  |      |  |
 |  | Additional: pointer     |    |      |   |      |      |      |  |
 |  | ----------------------- |    |      |   |      |      |      |  |
 |  | NextGroup: PSymbolGroup |    |      |   |      |      |      |  |
 |  |-----|-------------------|    |------|   |------|      |------|  |
 |--------|-----------------------------------------------------------|
          |
         \/
}

{$D+,L+,Y+}

unit DList;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  PSymbolGroup = ^TSymbolGroup;

  TSymbol = record
    Active: boolean;
    Structure: integer;
    Additional: pointer;
    NextGroup: PSymbolGroup;
  end;

  TSymbolGroup = record
    PrevGroup: PSymbolGroup;
    PrevChar: char;
    Chars: array[char] of TSymbol;
  end;

  THeaderAction = (haRead, haWrite);

  TDisposeStructure = procedure(Sender: TObject;
    Structure: Integer; Properties: Pointer) of Object;
  TFileHeader = procedure(Sender: TObject; H: integer;
    Action:THeaderAction; Format,InitPos: integer) of Object;
  TSaveStructure = procedure(Sender: TObject;
    Structure: Integer; Properties: Pointer; H: integer;
    Format: integer) of Object;
  TLoadStructure = procedure(Sender: TObject;
    Structure: Integer; var Properties: Pointer; H: integer;
    Format: integer) of Object;
  TFilter = procedure(Sender: TObject;
    Str: string; Structure: Integer; Properties: Pointer;
    var Enabled: boolean) of Object;
  TDeleteFilter = procedure(Sender: TObject;
    Str: string; Structure: Integer; Properties: Pointer;
    var CanDelete: boolean) of Object;

  TDList = class(Tcomponent)
  private
    { Private declarations }
    FSeparator: char;
    FOnDisposeStructure: TDisposeStructure;
    FOnFileHeader: TFileHeader;
    FOnSaveStructure: TSaveStructure;
    FOnLoadStructure: TLoadStructure;
    FOnFilter: TFilter;
    FOnDeleteFilter: TDeleteFilter;
    FirstGroup, ScrollGroup, BridgeGroup: PSymbolGroup;
    ScrollChar, BridgeChar: char;
    ScrollLen, RootLen, Level1, Level2, CurLevel, IncLevel: integer;
    ScrollStr, Root: string;
    function GetSymbol(Str: string; var c:char): PSymbolGroup;
    function GetPart(Str: string; var c:char): PSymbolGroup;
    procedure ResetSymbolGroup(SymbolGroup: PSymbolGroup);
    procedure ClearSymbolGroup(Str: string; var Group: PSymbolGroup; Level: integer);
    procedure ClearAllSymbolGroup(var Group: PSymbolGroup);
    function GoToNext: boolean;
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetFilter(RootStr: string; ForLevel, ToLevel: integer);
    procedure Add(Str: string; Struct: integer; Properties: pointer);
    function SetFields(Str: string; Struct: integer;
      Properties: pointer): boolean;
    function GetFields(Str: string; var Struct: integer;
      var Properties: pointer):boolean;
    procedure Remove(Str: string);
    procedure Clear;
    procedure ClearAll;
    function First: boolean;
    function Next: boolean;
    function GetName(var Str: string): boolean;
    function GetStructure(var Structure: integer): boolean;
    function GetProperties(var Properties: pointer): boolean;
    procedure Load(FileName: string; Format: integer);
    procedure Save(FileName: string; Format: integer);
    procedure GetList;
  published
    { Published declarations }
    property Separator:char read FSeparator write FSeparator stored true;
    property OnDisposeStructure: TDisposeStructure
      read FOnDisposeStructure write FOnDisposeStructure;
    property OnFileHeader: TFileHeader
      read FOnFileHeader write FOnFileHeader;
    property OnSaveStructure: TSaveStructure
      read FOnSaveStructure write FOnSaveStructure;
    property OnLoadStructure: TLoadStructure
      read FOnLoadStructure write FOnLoadStructure;
    property OnFilter: TFilter read FOnFilter write FOnFilter;
    property OnDeleteFilter: TDeleteFilter
      read FOnDeleteFilter write FOnDeleteFilter;
  end;

  TTracePos = record
    TraceGroup: PSymbolGroup;
    TraceChar: char;
  end;

  TTracer = class
  private
    { Private declarations }
    DList: TDList;
    CurGroup: PSymbolGroup;
    CurPos: TTracePos;
  public
    { Public declarations }
    constructor Create(DL: TDList);
    destructor Destroy; override;
    procedure SaveTracePos(var TracePos: TTracePos);
    procedure RestoreTracePos(TracePos: TTracePos);
    procedure BeginTrace(TraceRoot: string);
    function TraceNext(c: char; var Act: boolean; var Struct: integer;
      var Properties: pointer): boolean;
  end;

  procedure WriteString(H: Integer; s: string);
  procedure ReadString(H: Integer; var s: string);
  procedure WriteStrings(H: Integer; Lines: TStrings);
  procedure ReadStrings(H: Integer; var Lines: TStrings);

implementation

procedure WriteString(H: Integer; s: string);
var
  Len, i: integer;
begin
  Len:=Length(s);
  FileWrite(H, Len, SizeOf(Len));
  for i:=1 to Len do FileWrite(H, s[i], 1);
end;

procedure ReadString(H: Integer; var s: string);
var
  Len, i: integer;
begin
  FileRead(H, Len, SizeOf(Len));
  SetLength(s, Len);
  for i:=1 to Len do FileRead(H, s[i], 1);
end;

procedure WriteStrings(H: Integer; Lines: TStrings);
var
  LC, i: integer;
begin
  LC:=Lines.Count;
  FileWrite(H, LC, SizeOf(LC));
  for i:=0 to LC-1 do WriteString(H, Lines[i]);
end;

procedure ReadStrings(H: Integer; var Lines: TStrings);
var
  LC, i: integer;
  s: string;
begin
  FileRead(H, LC, SizeOf(LC));
  Lines.Clear;
  for i:=0 to LC - 1 do begin
    ReadString(H, s);
    Lines.Add(s);
  end;
end;

{TDList}
constructor TDList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FirstGroup:=nil;
  ScrollGroup:=nil;
  Separator := '_';
  SetFilter('',0,0);
end;

Destructor TDList.Destroy;
begin
  ClearAll;
  inherited destroy;
end;

procedure TDList.SetFilter(RootStr: string; ForLevel, ToLevel: integer);
begin
  Root:=RootStr;
  ScrollGroup:=nil;
  ScrollStr:='';
  RootLen:=Length(Root);
  Level1:=ForLevel;
  Level2:=ToLevel;
end;

procedure TDList.ResetSymbolGroup(SymbolGroup: PSymbolGroup);
var
  c:char;
begin
  for c:=Low(char) to High(char) do with SymbolGroup^.Chars[c] do begin
    Active:=false;
    Structure:=0;
    Additional:=nil;
    NextGroup:=nil;
  end;
end;

procedure TDList.Add(Str: string; Struct: integer; Properties: pointer);
var
  i:integer;
  c:char;
  CurGroup:PSymbolGroup;
begin
  if Str<>'' then begin
    CurGroup:=FirstGroup;
    if (CurGroup=nil)and(Length(Str)>0) then begin
      New(CurGroup);
      CurGroup^.PrevGroup:=nil;
      CurGroup^.PrevChar:=#0;
      FirstGroup:=CurGroup;
      ResetSymbolGroup(CurGroup);
    end;
    for i:=1 to Length(Str) do begin
      c:=Str[i];
      if i=Length(Str) then with CurGroup^.Chars[c] do begin
        try
          if Additional<>nil then if Assigned(FOnDisposeStructure) then
            FOnDisposeStructure(Self,Structure,Additional);
        finally
          Active:=true;
          Structure:=Struct;
          Additional:=Properties;
        end;
      end
      else begin
        if CurGroup^.Chars[c].NextGroup=nil then begin
          New(CurGroup^.Chars[c].NextGroup);
          ResetSymbolGroup(CurGroup^.Chars[c].NextGroup);
          CurGroup^.Chars[c].NextGroup^.PrevGroup:=CurGroup;
          CurGroup^.Chars[c].NextGroup^.PrevChar:=c;
        end;
        CurGroup:=CurGroup^.Chars[c].NextGroup;
      end;
    end;
  end;
end;

function TDList.GetSymbol(Str: string; var c: char): PSymbolGroup;
var
  CurGroup:PSymbolGroup;
  i,Len:integer;
begin
  Result:=nil;
  Len:=Length(Str);
  CurGroup:=FirstGroup;
  c:=#0;
  i:=1;
  while (CurGroup<>nil)and(i<=Len) do with CurGroup^.Chars[Str[i]] do begin
    if Active and (i=Len) then begin
      Result:=CurGroup;
      c:=Str[i];
      Exit;
    end;
    CurGroup:=NextGroup;
    inc(i);
  end;
end;

function TDList.GetPart(Str: string; var c: char): PSymbolGroup;
var
  CurGroup:PSymbolGroup;
  i,Len:integer;
begin
  Result:=nil;
  Len:=Length(Str);
  CurGroup:=FirstGroup;
  c:=#0;
  i:=1;
  while (CurGroup<>nil)and(i<=Len) do with CurGroup^.Chars[Str[i]] do begin
    if i=Len then begin
      Result:=CurGroup;
      c:=Str[i];
      Exit;
    end;
    CurGroup:=NextGroup;
    inc(i);
  end;
end;

function TDList.SetFields(Str: string; Struct: integer;
  Properties: pointer): boolean;
var
  Group: PSymbolGroup;
  c: char;
begin
  Result:=false;
  Group:=GetSymbol(Str, c);
  if Group<>nil then with Group^.Chars[c] do begin
    Result:=true;
    Structure:=Struct;
    if Additional<>nil then begin
      try
        if Assigned(FOnDisposeStructure) then
          FOnDisposeStructure(Self, Structure, Additional);
      finally
        Additional:=nil;
      end;
    end;
    Additional:=Properties;
  end;
end;

function TDList.GetFields(Str: string; var Struct: integer;
  var Properties: pointer): boolean;
var
  Group: PSymbolGroup;
  c: char;
begin
  Result:=false;
  Group:=GetSymbol(Str,c);
  if Group<>nil then with Group^.Chars[c] do begin
    Result:=true;
    Struct:=Structure;
    Properties:=Additional;
  end;
end;

procedure TDList.Remove(Str:string);
var
  CurGroup: PSymbolGroup;
  i: integer;
  c: char;
begin
  if Str<>'' then begin
    i:=1;
    CurGroup:=FirstGroup;
    if CurGroup <> nil then while true do begin
      if i>=Length(Str) then begin
        CurGroup^.Chars[Str[i]].Active:=false;
        with CurGroup^.Chars[Str[i]] do if Additional<>nil then begin
          try
            if Assigned(FOnDisposeStructure) then
              FOnDisposeStructure(Self,Structure,Additional);
          finally
            Additional:=nil;
          end;
        end;
        CurGroup^.Chars[Str[i]].Structure:=0;
        for i:=i downto 1 do begin
          for c:=Low(char) to High(Char) do if CurGroup^.Chars[c].Active or
            (CurGroup^.Chars[c].NextGroup<>nil) then Exit;
          if CurGroup=FirstGroup then begin
            if FirstGroup=ScrollGroup then ScrollGroup:=nil;
            DisPose(FirstGroup);
            FirstGroup:=nil;
            Exit;
          end
          else begin
            if CurGroup=ScrollGroup then ScrollGroup:=nil;
            c:=CurGroup^.PrevChar;
            CurGroup:=CurGroup^.PrevGroup;
            Dispose(CurGroup^.Chars[c].NextGroup);
            CurGroup^.Chars[c].NextGroup:=nil;
          end;
        end;
        Exit;
      end
      else begin
        CurGroup:=CurGroup^.Chars[Str[i]].NextGroup;
        if CurGroup=nil then Exit;
        inc(i);
      end;
    end;
  end;
end;

procedure TDList.ClearSymbolGroup(Str: string; var Group: PSymbolGroup; Level: integer);
var
  c: char;
  AllPassive, CanDel: boolean;
begin
  if Group<>nil then begin
    AllPassive:=true;
    for c:=Low(Char) to High(Char) do with Group^.Chars[c] do begin
      if c=Separator then IncLevel:=1 else IncLevel:=0;
      if (NextGroup<>nil)and(Level+IncLevel<=Level2) then
        ClearSymbolGroup(Str+c,NextGroup,Level+IncLevel);
      if (Level+IncLevel>=Level1)and(Level+IncLevel<=Level2) then begin
        if Active then begin
          CanDel:=true;
          if Assigned(FOnDeleteFilter) then
            FOnDeleteFilter(Self, Str+c, Structure, Additional, CanDel);
          if CanDel then begin
            Active:=false;
            try
              if Additional<>nil then if Assigned(FOnDisposeStructure) then
                FOnDisposeStructure(Self,Structure,Additional);
            finally
              Structure:=0;
              Additional:=nil;
            end;
          end
          else AllPassive:=false;
        end;
        if AllPassive then if NextGroup<>nil then AllPassive:=false;
      end
      else AllPassive:=false;
    end;
    if AllPassive then begin
      if ScrollGroup=Group then ScrollGroup:=nil;
      Dispose(Group);
      Group:=nil;
    end;
  end;
end;

procedure TDList.Clear;
var
  Group:PSymbolGroup;
  c:char;
begin
  if Root='' then ClearSymbolGroup('',FirstGroup,0)
  else begin
    Group:=GetPart(Root,c);
    if Group<>nil then ClearSymbolGroup(Root,Group^.Chars[c].NextGroup,0);
  end;
end;

procedure TDList.ClearAllSymbolGroup(var Group:PSymbolGroup);
var
  c:char;
begin
  if Group<>nil then begin
    for c:=Low(Char) to High(Char) do with Group^.Chars[c] do begin
      if NextGroup<>nil then ClearAllSymbolGroup(NextGroup);
      if Additional<>nil then begin
        try
          if Assigned(FOnDisposeStructure) then
            FOnDisposeStructure(Self,Structure,Additional);
        finally
          Additional:=nil;
        end;
      end;
    end;
    if ScrollGroup=Group then ScrollGroup:=nil;
    Dispose(Group);
    Group:=nil;
  end;
end;

procedure TDList.ClearAll;
begin
  ClearAllSymbolGroup(FirstGroup);
end;

function TDList.First:boolean;
var
  Group:PSymbolGroup;
  c:char;
  s: string;
  St: integer;
  p: pointer;
  Enab: boolean;
begin
  Result:=false;
  if Root='' then Group:=FirstGroup
  else begin
    Group:=GetPart(Root,c);
    if Group=nil then Exit
    else Group:=Group^.Chars[c].NextGroup;
  end;
  ScrollGroup:=Group;
  if Group=nil then Exit
  else begin
    ScrollChar:=#0;
    ScrollStr:=ScrollChar;
    ScrollLen:=RootLen;
    CurLevel:=0;
    IncLevel:=0;
    if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
    if ScrollGroup^.Chars[ScrollChar].Active and
      (CurLevel+IncLevel>=Level1) and (CurLevel+IncLevel<=Level2) then begin
      Enab:=true;
      if Assigned(FOnFilter) then
        if GetName(s) and GetStructure(St) and GetProperties(p) then
          FOnFilter(Self, s, St, p, Enab)
        else Enab:=false;
      if Enab then begin
        Result:=true;
        Exit;
      end;
    end;
    Result:=Next;
  end;
end;

function TDList.GoToNext: boolean;
begin
  BridgeGroup:=ScrollGroup;
  BridgeChar:=ScrollChar;
  Result:=false;
  try
    if ScrollGroup<>nil then repeat
      while (ScrollGroup^.Chars[ScrollChar].NextGroup<>nil)and
        ((Level2>CurLevel)or(ScrollChar<>Separator)) do begin
        ScrollGroup:=ScrollGroup^.Chars[ScrollChar].NextGroup;
        if ScrollChar=Separator then inc(CurLevel);
        ScrollChar:=#0;
        ScrollStr:=ScrollStr+#0;
        inc(ScrollLen);
        if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
        if ScrollGroup^.Chars[ScrollChar].Active and
          (CurLevel+IncLevel>=Level1)and(CurLevel+IncLevel<=Level2) then begin
            Result:=true;
            Exit;
          end;
      end;
      while ScrollChar=High(char) do if RootLen=ScrollLen then Exit
      else begin
        ScrollChar:=ScrollGroup^.PrevChar;
        ScrollGroup:=ScrollGroup^.PrevGroup;
        ScrollStr:=Copy(ScrollStr,1,Length(ScrollStr)-1);
        dec(ScrollLen);
        if ScrollChar=Separator then dec(CurLevel);
      end;
      inc(ScrollChar);
      if ScrollChar=Separator then IncLevel:=1 else IncLevel:=0;
      ScrollStr[Length(ScrollStr)]:=ScrollChar;
    until ScrollGroup^.Chars[ScrollChar].Active and
      (CurLevel+IncLevel>=Level1)and(CurLevel+IncLevel<=Level2)
    else Exit;
    Result:=true;
  finally
    if not Result then begin
      ScrollGroup:=BridgeGroup;
      ScrollChar:=BridgeChar;
    end;
  end;
end;

function TDList.Next: boolean;
var
  s: string;
  St: integer;
  p: pointer;
  Enab: boolean;
begin
  if Assigned(FOnFilter) then begin
    repeat
      if GoToNext then begin
        Enab:=true;
        if GetName(s) and GetStructure(St) and GetProperties(p) then
          FOnFilter(Self, s, St, p, Enab)
        else Enab:=false;
      end
      else begin
        Result:=false;
        Exit;
      end;
    until Enab;
    Result:=true;
  end
  else Result:=GoToNext;
end;

function TDList.GetName(var Str: string):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Str:=Root+ScrollStr;
    Result:=true;
  end;
end;

function TDList.GetStructure(var Structure:integer):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Structure:=ScrollGroup^.Chars[ScrollChar].Structure;
    Result:=true;
  end;
end;

function TDList.GetProperties(var Properties: pointer):boolean;
begin
  if ScrollGroup=nil then Result:=false
  else begin
    Properties:=ScrollGroup^.Chars[ScrollChar].Additional;
    Result:=true;
  end;
end;

procedure TDList.Save(FileName:string; Format:integer);
var
  BackupName, Str: string;
  H, Structure, i, Len: integer;
  Properties: pointer;
begin
  if FileExists(FileName) then begin {Creation of a backup file}
    BackupName:=ExtractFilePath(FileName)+'Backup of '+
      ExtractFileName(FileName);
    DeleteFile(BackupName);
    RenameFile(FileName, BackupName);
  end;
  H := FileCreate(FileName);
  if H > -1 then begin
    try
      FileWrite(H, Format, SizeOf(Format));
      if Assigned(FOnFileHeader) then
        FOnFileHeader(Self, H, haWrite, Format, SizeOf(Format));
      if First then repeat
        GetName(Str);
        Len:=Length(Str);
        FileWrite(H, Len, SizeOf(Len));
        for i:=1 to Len do FileWrite(H, Str[i], 1);
        GetStructure(Structure);
        FileWrite(H, Structure, SizeOf(Structure));
        GetProperties(Properties);
        if Assigned(FOnSaveStructure) then
          FOnSaveStructure(Self, Structure, Properties, H, Format);
      until not Next;
    finally
      FileClose(H);
    end;
  end;
end;

procedure TDList.Load(FileName:string; Format:integer);
var
  Str: string;
  H, ReadingFormat, Structure, i, Len: integer;
  Properties: pointer;
begin
{const fmOpenRead       = $0000;
const fmOpenWrite      = $0001;
const fmOpenReadWrite  = $0002;
const fmShareCompat    = $0000;
const fmShareExclusive = $0010;
const fmShareDenyWrite = $0020;
const fmShareDenyRead  = $0030;
const fmShareDenyNone  = $0040;}
  H := FileOpen(FileName, fmOpenRead);
  if H > -1 then begin
    try
      FileRead(H, ReadingFormat, SizeOf(Format));
      if ReadingFormat <= Format then begin
        if Assigned(FOnFileHeader) then
          FOnFileHeader(Self, H, haRead, ReadingFormat, SizeOf(ReadingFormat));
        while true do begin
          if FileRead(H, Len, SizeOf(Len)) = 0 then Exit;
          SetLength(Str, Len);
          for i:=1 to Len do FileRead(H, Str[i], 1);
          FileRead(H, Structure, SizeOf(Structure));
          Properties:=nil;
          if Assigned(FOnLoadStructure) then
            FOnLoadStructure(Self, Structure, Properties, H, ReadingFormat);
          Add(Str, Structure, Properties);
        end;
      end;
    finally
      FileClose(H);
    end;
  end;
end;

procedure TDList.GetList;
begin
  if First then repeat
  until not Next;
end;

{TTracer}
constructor TTracer.Create(DL: TDList);
begin
  inherited Create;
  DList:=DL;
end;

destructor TTracer.Destroy;
begin
  inherited Destroy;
end;

procedure TTracer.BeginTrace(TraceRoot: string);
begin
  if TraceRoot='' then CurPos.TraceGroup:=nil
  else CurPos.TraceGroup:=DList.GetPart(TraceRoot, CurPos.TraceChar);
end;

procedure TTracer.SaveTracePos(var TracePos: TTracePos);
begin
  TracePos:=CurPos;
end;

procedure TTracer.RestoreTracePos(TracePos: TTracePos);
begin
  CurPos:=TracePos;
end;

function TTracer.TraceNext(c: char; var Act: boolean; var Struct: integer;
  var Properties: pointer): boolean;
begin
  if CurPos.TraceGroup=nil then CurGroup:=DList.FirstGroup
  else CurGroup:=CurPos.TraceGroup^.Chars[CurPos.TraceChar].NextGroup;
  if CurGroup=nil then Result:=false
  else with CurGroup^.Chars[c] do if Active or (NextGroup<>nil) then begin
    CurPos.TraceGroup:=CurGroup;
    CurPos.TraceChar:=c;
    Act:=Active;
    Struct:=Structure;
    if Act then Properties:=Additional;
    Result:=true;
  end
  else Result:=false;
end;

end.

